""" tlog environment configuration tests """
import os
import stat
import socket
import inspect

from shutil import copyfile
from tempfile import mkdtemp
import pytest

from misc import check_recording, mklogfile, ssh_pexpect
from config import TlogRecConfig, TlogRecSessionConfig


TLOG_REC_SESSION_PROG = "/usr/bin/tlog-rec-session"
TLOG_REC_PROG = "/usr/bin/tlog-rec"
SYSTEM_TLOG_REC_SESSION_CONF = "/etc/tlog/tlog-rec-session.conf"
SYSTEM_TLOG_REC_CONF = "/etc/tlog/tlog-rec.conf"
TMP_TLOG_REC_SESSION_CONF = "/tmp/tlog-rec-session.conf-test"
TMP_TLOG_REC_CONF = "/tmp/tlog-rec.conf-test"
TEXT_ENV_TLOG_REC_SESSION_CONF = "/tmp/tlog-rec-session.conf-text-test"
TEXT_ENV_TLOG_REC_CONF = "/tmp/tlog-rec.conf-text-test"
TLOG_TEST_LOCAL_USER = "tlitestlocaluser1"
TLOG_TEST_LOCAL_ADMIN = "tlitestlocaladmin1"


@pytest.fixture(scope="module")
def session_env_config_setup():
    """ Setup/teardown fixture applied to
    TestTlogRecSession tests"""
    conf_file = SYSTEM_TLOG_REC_SESSION_CONF
    backup_file = f"{conf_file}-orig"
    copyfile(conf_file, backup_file)
    yield session_env_config_setup
    # restore original configuration
    copyfile(backup_file, conf_file)
    os.remove(backup_file)


class TestTlogRecSession:
    """ tlog-rec-session tests """
    orig_hostname = socket.gethostname()
    tempdir = mkdtemp(prefix='/tmp/TestTlogRecSession.')
    user1 = TLOG_TEST_LOCAL_USER
    admin1 = TLOG_TEST_LOCAL_ADMIN
    os.chmod(tempdir, stat.S_IRWXU + stat.S_IRWXG + stat.S_IRWXO +
             stat.S_ISUID + stat.S_ISGID + stat.S_ISVTX)

    def test_conf_file_var(self, session_env_config_setup):
        """ Validate configuration settings are overwritten
        by TLOG_REC_SESSION_CONF_FILE variable
        """
        logfile = mklogfile(self.tempdir)
        msg = inspect.stack()[0][3]
        input_notice = "Test NOTICE"
        tmp_conf_file = TMP_TLOG_REC_SESSION_CONF

        # system wide journal configuration file
        sessionclass_system = TlogRecSessionConfig(writer="journal")
        sessionclass_system.generate_config(SYSTEM_TLOG_REC_SESSION_CONF)

        # temporary configuration file to override with
        sessionclass_tmp = TlogRecSessionConfig(writer="file",
                                                file_writer_path=logfile,
                                                notice=input_notice)
        sessionclass_tmp.generate_config(tmp_conf_file)

        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        shell.sendline(f'export TLOG_REC_SESSION_CONF_FILE={tmp_conf_file}')
        shell.sendline(TLOG_REC_SESSION_PROG)
        # validate the notice override
        shell.expect(input_notice)
        shell.sendline(f'echo {msg}')
        shell.expect(msg)
        shell.sendline('exit')
        check_recording(shell, msg, logfile)
        shell.close()

    def test_conf_text_var(self, session_env_config_setup):
        """ Validate configuration settings are overwritten
        by TLOG_REC_SESSION_CONF_TEXT variable
        """
        logfile = mklogfile(self.tempdir)
        msg = inspect.stack()[0][3]
        tmp_notice_msg = "temp"
        tmp_conf_file = TMP_TLOG_REC_SESSION_CONF

        # system wide configuration file
        sessionclass_system = TlogRecSessionConfig(writer="journal")
        sessionclass_system.generate_config(SYSTEM_TLOG_REC_SESSION_CONF)

        # temporary configuration file
        sessionclass_tmp = TlogRecSessionConfig(writer="file",
                                                file_writer_path=logfile,
                                                notice=tmp_notice_msg)
        sessionclass_tmp.generate_config(tmp_conf_file)

        # validate the notice override
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        shell.sendline(f'export TLOG_REC_SESSION_CONF_FILE={tmp_conf_file}')
        shell.sendline('export TLOG_REC_SESSION_CONF_TEXT='
                       '\'{"notice":"TEXT Notice"}\'')
        shell.sendline(TLOG_REC_SESSION_PROG)
        shell.expect("TEXT Notice")
        shell.sendline(f'echo {msg}')
        shell.expect(msg)
        shell.sendline('exit')
        shell.close()

    def test_conf_shell_var(self, session_env_config_setup):
        """ Validate the TLOG_REC_SESSION_SHELL variable
        is honored
        """
        msg = inspect.stack()[0][3]
        input_shell = "/bin/tcsh"

        sessionclass = TlogRecSessionConfig(writer="journal",
                                            shell='/bin/bash')
        sessionclass.generate_config(SYSTEM_TLOG_REC_SESSION_CONF)

        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        shell.sendline('export TLOG_REC_SESSION_CONF_TEXT='
                       '\'{"shell":"/bin/zsh"}\'')
        shell.sendline(f'export TLOG_REC_SESSION_SHELL={input_shell}')
        shell.sendline(TLOG_REC_SESSION_PROG)
        shell.expect("ATTENTION")
        shell.sendline("echo $SHELL")
        shell.expect(input_shell)
        shell.sendline(f'echo {msg}')
        shell.expect(msg)
        shell.sendline('exit')
        shell.close()


@pytest.fixture(scope="module")
def rec_env_config_setup():
    """ Setup/teardown fixture applied to
    TestTlogRec tests"""
    conf_file = SYSTEM_TLOG_REC_CONF
    backup_file = f"{conf_file}-orig"
    copyfile(conf_file, backup_file)
    yield rec_env_config_setup
    # restore original configuration
    copyfile(backup_file, conf_file)
    os.remove(backup_file)


class TestTlogRec:
    """ tlog-rec tests """
    orig_hostname = socket.gethostname()
    tempdir = mkdtemp(prefix='/tmp/TestTlogRec.')
    user1 = TLOG_TEST_LOCAL_USER
    admin1 = TLOG_TEST_LOCAL_ADMIN
    os.chmod(tempdir, stat.S_IRWXU + stat.S_IRWXG + stat.S_IRWXO +
             stat.S_ISUID + stat.S_ISGID + stat.S_ISVTX)

    def test_conf_file_var(self, rec_env_config_setup):
        """ Validate configuration settings are overwritten
        by TLOG_REC_CONF_FILE variable
        """
        logfile = mklogfile(self.tempdir)
        msg = inspect.stack()[0][3]
        tmp_conf_file = TMP_TLOG_REC_CONF

        # system wide configuration file
        recclass_system = TlogRecConfig(writer="journal")
        recclass_system.generate_config(SYSTEM_TLOG_REC_CONF)

        # temporary configuration file to override with
        recclass_tmp = TlogRecConfig(writer="file", file_writer_path=logfile)
        recclass_tmp.generate_config(tmp_conf_file)

        # validate the file writer override
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        shell.sendline(f'export TLOG_REC_CONF_FILE={tmp_conf_file}')
        shell.sendline(TLOG_REC_PROG + f"-c echo {msg}")
        check_recording(shell, msg, logfile)
        shell.close()

    def test_conf_text_var(self, rec_env_config_setup):
        """ Validate configuration settings are overwritten
        by TLOG_REC_CONF_TEXT variable
        """
        logfile = mklogfile(self.tempdir)
        msg = inspect.stack()[0][3]
        tmp_conf_file = TMP_TLOG_REC_CONF

        # system wide configuration file
        recclass_system = TlogRecConfig(writer="journal")
        recclass_system.generate_config(SYSTEM_TLOG_REC_CONF)

        # temporary configuration file
        recclass_tmp = TlogRecSessionConfig(writer="syslog")
        recclass_tmp.generate_config(tmp_conf_file)

        # validate the file writer override
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        shell.sendline(f'export TLOG_REC_CONF_FILE={tmp_conf_file}')
        shell.sendline('export TLOG_REC_CONF_TEXT=\'{"writer":"file"}\'')
        shell.sendline(TLOG_REC_PROG + f"-c echo {msg}")
        check_recording(shell, msg, logfile)
        shell.close()

    def test_conf_shell_var(self, session_env_config_setup):
        """ Validate the SHELL variable is honored when no
        command to record was specified
        """
        input_shell = "/bin/tcsh"

        recclass = TlogRecConfig(writer="journal")
        recclass.generate_config(SYSTEM_TLOG_REC_CONF)

        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        shell.sendline(f'export SHELL={input_shell}')
        shell.sendline(TLOG_REC_PROG)
        shell.sendline("echo $SHELL")
        shell.expect(input_shell)
        shell.sendline('exit')
        shell.close()



""" tlog tests """
import os
import stat
import time
import inspect
from tempfile import mkdtemp
from subprocess import Popen, PIPE, STDOUT
import pytest

from misc import check_recording, mklogfile, mkcfgfile, \
                 ssh_pexpect, check_recording_missing, copyfile
from config import TlogRecConfig, TlogRecSessionConfig

TLOG_REC_SESSION_PROG = "/usr/bin/tlog-rec-session"
SYSTEM_TLOG_REC_SESSION_CONF = "/etc/tlog/tlog-rec-session.conf"

@pytest.fixture
def utempter_enabled():
    p = Popen(['ldd', TLOG_REC_SESSION_PROG],
              stdout=PIPE, stdin=PIPE, stderr=PIPE, encoding='utf8')
    stdout_data = p.communicate()[0]
    return 'libutempter.so' in stdout_data


class TestTlogRecSession:
    """ Test tlog-rec-session functionality """
    user = 'tlitestlocaluser2'
    tempdir = mkdtemp(prefix='/tmp/TestRecSession.')
    os.chmod(tempdir, stat.S_IRWXU + stat.S_IRWXG + stat.S_IRWXO +
             stat.S_ISUID + stat.S_ISGID + stat.S_ISVTX)

    @pytest.mark.tier1
    def test_session_record_to_file(self):
        """
        Check tlog-rec-session preserves session in a file
        """
        myname = inspect.stack()[0][3]
        logfile = mklogfile(self.tempdir)
        sessionclass = TlogRecSessionConfig(writer="file",
                                            file_writer_path=logfile)
        sessionclass.generate_config(SYSTEM_TLOG_REC_SESSION_CONF)
        shell = ssh_pexpect(self.user, 'Secret123', 'localhost')
        shell.sendline('echo {}'.format(myname))
        shell.sendline('exit')
        check_recording(shell, myname, logfile)
        shell.close()

    @pytest.mark.tier1
    def test_session_record_to_journal(self):
        """
        Check tlog-rec-session preserves session in journal
        """
        myname = inspect.stack()[0][3]
        sessionclass = TlogRecSessionConfig(writer="journal")
        sessionclass.generate_config(SYSTEM_TLOG_REC_SESSION_CONF)
        shell = ssh_pexpect(self.user, 'Secret123', 'localhost')
        shell.sendline('echo {}'.format(myname))
        shell.sendline('exit')
        check_recording(shell, myname)
        shell.close()

    @pytest.mark.tier1
    def test_session_record_to_syslog(self):
        """
        Check tlog-rec-session preserves session via syslog
        """
        myname = inspect.stack()[0][3]
        sessionclass = TlogRecSessionConfig(writer="syslog")
        sessionclass.generate_config(SYSTEM_TLOG_REC_SESSION_CONF)
        shell = ssh_pexpect(self.user, 'Secret123', 'localhost')
        shell.sendline('echo {}'.format(myname))
        shell.sendline('exit')
        check_recording(shell, myname)
        shell.close()

    def test_session_record_fast_input_with_latency(self):
        """
        Check tlog-rec-session caches data some time before logging
        """
        myname = inspect.stack()[0][3]
        logfile = mklogfile(self.tempdir)
        sessionclass = TlogRecSessionConfig(writer="file",
                                            file_writer_path=logfile,
                                            latency=15)
        shell = ssh_pexpect(self.user, 'Secret123', 'localhost')
        for num in range(0, 200):
            shell.sendline('echo {}_{}'.format(myname, num))
        shell.sendline('exit')
        check_recording(shell, '{}_199'.format(myname), logfile)
        shell.close()

    def test_session_record_fast_input_with_payload(self):
        """
        Check tlog-rec limits output payload size
        """
        myname = inspect.stack()[0][3]
        logfile = mklogfile(self.tempdir)
        sessionclass = TlogRecSessionConfig(writer="file",
                                            file_writer_path=logfile,
                                            payload=128)
        sessionclass.generate_config(SYSTEM_TLOG_REC_SESSION_CONF)
        shell = ssh_pexpect(self.user, 'Secret123', 'localhost')
        for num in range(0, 200):
            shell.sendline('echo {}_{}'.format(myname, num))
        shell.sendline('exit')
        check_recording(shell, '{}_199'.format(myname), logfile)
        shell.close()

    def test_session_record_fast_input_with_limit_rate(self):
        """
        Check tlog-rec-session records session with limit rate
        configured
        """
        myname = inspect.stack()[0][3]
        logfile = mklogfile(self.tempdir)
        sessionclass = TlogRecSessionConfig(writer="file",
                                            file_writer_path=logfile,
                                            limit_rate=10)
        sessionclass.generate_config(SYSTEM_TLOG_REC_SESSION_CONF)
        shell = ssh_pexpect(self.user, 'Secret123', 'localhost')
        for num in range(0, 200):
            shell.sendline('echo {}_{}'.format(myname, num))
        shell.sendline('exit')
        check_recording(shell, '{}_199'.format(myname), logfile)
        shell.close()

    def test_session_record_fast_input_with_limit_burst(self):
        """
        Check tlog-rec-session allows limited burst of fast output
        """
        myname = inspect.stack()[0][3]
        logfile = mklogfile(self.tempdir)
        sessionclass = TlogRecSessionConfig(writer="file",
                                            file_writer_path=logfile,
                                            limit_rate=10,
                                            limit_burst=100)
        sessionclass.generate_config(SYSTEM_TLOG_REC_SESSION_CONF)
        shell = ssh_pexpect(self.user, 'Secret123', 'localhost')
        for num in range(0, 200):
            shell.sendline('echo {}_{}'.format(myname, num))
        shell.sendline('exit')
        check_recording(shell, '{}_199'.format(myname), logfile)
        shell.close()

    def test_session_record_fast_input_with_limit_action_drop(self):
        """
        Check tlog-rec-session drops output when logging limit reached
        """
        logfile = mklogfile(self.tempdir)
        sessionclass = TlogRecSessionConfig(writer="file",
                                            file_writer_path=logfile,
                                            limit_rate=10,
                                            limit_action="drop")
        sessionclass.generate_config(SYSTEM_TLOG_REC_SESSION_CONF)
        shell = ssh_pexpect(self.user, 'Secret123', 'localhost')
        shell.sendline('cat /usr/share/dict/linux.words')
        time.sleep(1)
        shell.sendline('exit')
        shell.close()
        shell = ssh_pexpect(self.user, 'Secret123', 'localhost')
        check_recording_missing(shell, 'Byronite', logfile)
        check_recording_missing(shell, 'zygote', logfile)

    def test_session_record_fast_input_with_limit_action_delay(self):
        """
        Check tlog-rec-session delays recording when logging limit reached
        """
        myname = inspect.stack()[0][3]
        logfile = mklogfile(self.tempdir)
        sessionclass = TlogRecSessionConfig(writer="file",
                                            file_writer_path=logfile,
                                            limit_rate=500,
                                            limit_action="delay")
        sessionclass.generate_config(SYSTEM_TLOG_REC_SESSION_CONF)
        shell = ssh_pexpect(self.user, 'Secret123', 'localhost')
        for num in range(0, 200):
            shell.sendline('echo {}_{}'.format(myname, num))
        shell.sendline('exit')
        check_recording(shell, '{}_199'.format(myname), logfile)
        shell.close()

    def test_session_record_fast_input_with_limit_action_pass(self):
        """
        Check tlog-rec-session ignores logging limits
        """
        myname = inspect.stack()[0][3]
        logfile = mklogfile(self.tempdir)
        sessionclass = TlogRecSessionConfig(writer="file",
                                            file_writer_path=logfile,
                                            limit_rate=500,
                                            limit_action="pass")
        sessionclass.generate_config(SYSTEM_TLOG_REC_SESSION_CONF)
        shell = ssh_pexpect(self.user, 'Secret123', 'localhost')
        for num in range(0, 200):
            shell.sendline('echo {}_{}'.format(myname, num))
        shell.sendline('exit')
        check_recording(shell, '{}_199'.format(myname), logfile)
        shell.close()

    def test_session_record_with_different_shell(self):
        """
        Check tlog-rec-session can specify different shell
        """
        logfile = mklogfile(self.tempdir)
        sessionclass = TlogRecSessionConfig(shell="/usr/bin/tcsh",
                                            writer="file",
                                            file_writer_path=logfile)
        sessionclass.generate_config(SYSTEM_TLOG_REC_SESSION_CONF)
        shell = ssh_pexpect(self.user, 'Secret123', 'localhost')
        shell.sendline('echo $SHELL')
        check_recording(shell, '/usr/bin/tcsh', logfile)
        shell.sendline('exit')

    def test_session_record_pipe_io_stdin(self):
        """
        Pipe I/O through stdin
        """
        text_in_stdio = 'print("hello world")\n'
        text_out = "hello world"
        p = Popen(['sshpass', '-p', 'Secret123', 'ssh', '-o',
                   'StrictHostKeyChecking=no',
                   'tlitestlocaluser2@localhost', 'python3'],
        stdout=PIPE, stdin=PIPE, stderr=PIPE, encoding='utf8')
        stdout_data = p.communicate(input=text_in_stdio)[0]
        assert text_out in stdout_data

    def test_session_record_user_in_utmp(self, utempter_enabled):
        """
        Check tlog-rec-session preserves session in a file
        """
        if not utempter_enabled:
            pytest.skip('utempter not enabled, skipping test')
        myname = inspect.stack()[0][3]
        whoami = '{} pts'.format(self.user)
        logfile = mklogfile(self.tempdir)
        sessionclass = TlogRecSessionConfig(writer="file",
                                            file_writer_path=logfile)
        sessionclass.generate_config(SYSTEM_TLOG_REC_SESSION_CONF)
        shell = ssh_pexpect(self.user, 'Secret123', 'localhost')
        shell.sendline('echo {}'.format(myname))
        shell.sendline('who am i')
        shell.sendline('exit')
        check_recording(shell, myname, logfile)
        check_recording(shell, whoami, logfile)
        shell.close()

    @classmethod
    def teardown_class(cls):
        """ Copy original conf file back into place """
        filename = SYSTEM_TLOG_REC_SESSION_CONF
        bkup = '{}.origtest'.format(filename)
        copyfile(bkup, filename)



""" tlog tests """
import os
import ast
import stat
import time
import socket
from tempfile import mkdtemp

import pexpect

import pytest

from misc import check_recording, ssh_pexpect, mklogfile, \
                 check_outfile, check_journal, mkcfgfile, \
                 journal_find_last


class TestTlogRec:
    """ tlog-rec tests """
    orig_hostname = socket.gethostname()
    tempdir = mkdtemp(prefix='/tmp/TestTlogRec.')
    user1 = 'tlitestlocaluser1'
    admin1 = 'tlitestlocaladmin1'
    os.chmod(tempdir, stat.S_IRWXU + stat.S_IRWXG + stat.S_IRWXO +
             stat.S_ISUID + stat.S_ISGID + stat.S_ISVTX)

    @pytest.mark.tier1
    def test_record_command_to_file(self):
        """
        Check tlog-rec preserves output when reording to file
        """
        logfile = mklogfile(self.tempdir)
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        shell.sendline('tlog-rec -o {} whoami'.format(logfile))
        check_outfile('out_txt\":\"{}'.format(self.user1), logfile)
        check_recording(shell, self.user1, logfile)
        shell.close()

    @pytest.mark.tier1
    def test_record_command_to_journal(self):
        """
        Check tlog-rec preserves output when recording to journal
        """
        match_filter = '_COMM=tlog-rec'
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        reccmd = 'test_record_to_journal'
        shell.sendline('tlog-rec -w journal echo {}'.format(reccmd))
        check_journal(reccmd)
        check_recording(shell, reccmd)
        shell.close()

    @pytest.mark.tier1
    def test_record_command_to_syslog(self):
        """
        Check tlog-rec preserves output when recording to syslog
        """
        match_filter = '_COMM=tlog-rec'
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        reccmd = 'test_record_to_syslog'
        shell.sendline('tlog-rec --writer=syslog echo {}'.format(reccmd))
        check_journal(reccmd)
        check_recording(shell, reccmd)
        shell.close()

    def test_record_interactive_session(self):
        """
        Check tlog-rec preserves activity during interactive
        session in recordings
        """
        logfile = mklogfile(self.tempdir)
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        shell.sendline('tlog-rec -o {}'.format(logfile))
        shell.sendline('whoami')
        shell.expect(self.user1)
        shell.sendline('sleep 2')
        shell.sendline('echo test123')
        shell.expect('test123')
        shell.sendline('echo test1123out>/tmp/pexpect.test1123out')
        check_outfile('test1123out', logfile)
        check_recording(shell, 'test1123out', logfile)
        shell.close()

    def test_record_binary_output(self):
        """
        Check tlog-rec preserves binary output in recordings
        """
        logfile = mklogfile(self.tempdir)
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        reccmd = 'cat /usr/bin/gzip'
        shell.sendline('tlog-rec -o {} {}'.format(logfile, reccmd))
        shell.expect(r'\u0000')
        check_recording(shell, r'\u0000', logfile)
        shell.close()

    def test_record_diff_char_sets(self):
        """
        Check tlog-rec preserves non-English I/O in recordings
        """
        logfile = '{}-ru_RU'.format(mklogfile(self.tempdir))
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        shell.sendline('tlog-rec -o {} /bin/bash'.format(logfile))
        shell.sendline('LANG=ru_RU.utf8 locale -ck LC_TIME')
        shell.sendline('exit')
        check_outfile('Январь', logfile)
        check_recording(shell, 'Январь', logfile)
        shell.close()

        logfile = '{}-el_GR'.format(mklogfile(self.tempdir))
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        shell.sendline('tlog-rec -o {} /bin/bash'.format(logfile))
        shell.sendline('LANG=el_GR.utf8 locale -ck LC_TIME')
        shell.sendline('exit')
        check_outfile('Ιανουάριος', logfile)
        check_recording(shell, 'Ιανουάριος', logfile)
        shell.close()

        logfile = '{}-en_US'.format(mklogfile(self.tempdir))
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        shell.sendline('tlog-rec -o {} /bin/bash'.format(logfile))
        shell.sendline('export LANG=en_US.utf8')
        shell.sendline('echo Watérmân')
        shell.sendline('exit')
        check_outfile('Watérmân', logfile)
        check_recording(shell, 'Watérmân', logfile)
        shell.expect('Watérmân')
        shell.close()

    def test_record_fast_input(self):
        """
        Check tlog-rec preserves fast flooded I/O in recordings
        """
        logfile = mklogfile(self.tempdir)
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        shell.sendline('tlog-rec -o {} /bin/bash'.format(logfile))
        for num in range(0, 2000):
            shell.sendline('echo test_{}'.format(num))
        shell.sendline('exit')
        for num in range(0, 2000, 100):
            check_recording(shell, 'test_{}'.format(num), logfile)
        shell.close()

    def test_record_as_unprivileged_user(self):
        """
        Check tlog-rec preserves unauthorized activity of
        unprivileged user in recordings
        """
        logfile = mklogfile(self.tempdir)
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')

        shell.sendline('tlog-rec -o {}'.format(logfile))
        shell.sendline('whoami')
        shell.expect(self.user1)
        shell.sendline('echo test1123out')
        shell.sendline('sleep 2')
        shell.sendline('ls -ltr /var/log/audit')
        check_recording(shell, 'test1123out', logfile)
        check_recording(shell, 'Permission denied', logfile)
        shell.sendline('exit')
        shell.close()

    def test_record_as_admin_user(self):
        """
        Check tlog-rec preserves sudo activity of admin user in
        recordings
        """
        logfile = mklogfile(self.tempdir)
        cfg = '''
        %wheel        ALL=(ALL)       NOPASSWD: ALL
        '''
        mkcfgfile('/etc/sudoers.d/01_wheel_nopass', cfg)
        shell = ssh_pexpect(self.admin1, 'Secret123', 'localhost')
        shell.sendline('tlog-rec -o {}'.format(logfile))
        shell.sendline('whoami')
        shell.expect(self.admin1)
        shell.sendline('sleep 2')
        shell.sendline('echo test1223')
        shell.expect('test1223')
        shell.sendline('sudo cd /root')
        shell.sendline('exit')
        check_outfile('test1223', logfile)
        check_recording(shell, 'test1223', logfile)
        shell.close()
        shell = ssh_pexpect(self.admin1, 'Secret123', 'localhost')
        check_recording(shell, 'cd /root', logfile)
        shell.close()

    # diable no-self-use in this function.  Otherwise pylint
    # complains that this method coule be a function.
    #
    # pylint: disable=no-self-use
    def test_record_from_different_hostnames(self):
        """
        Check tlog-rec reflects hostname changes in recordings

        This is to simulate receiving remote journal sessions
        """
        oldname = socket.gethostname()
        shell = pexpect.spawn('/bin/bash')
        for num in range(0, 3):
            newname = 'test{}-{}'.format(num, oldname)
            socket.sethostname(newname)
            open('/etc/hostname', 'w').write(newname)
            shell.sendline('hostname')
            shell.expect(newname)
            time.sleep(1)
            shell.sendline('tlog-rec -w journal whoami')
            time.sleep(1)
            shell.sendline('hostnamectl status')
            time.sleep(1)
            entry = journal_find_last()
            message = entry['MESSAGE']
            mhostname = ast.literal_eval(message)['host']
            assert mhostname == newname
            time.sleep(1)
        socket.sethostname(oldname)
        open('/etc/hostname', 'w').write(oldname)



"""Configuration module used for integration tests"""
import json
from misc import mkcfgfile

DEFAULT_TLOG_REC_LATENCY = 10
DEFAULT_TLOG_REC_PAYLOAD = 2048
DEFAULT_TLOG_REC_LOG_INPUT = False
DEFAULT_TLOG_REC_LOG_OUTPUT = True
DEFAULT_TLOG_REC_LOG_WINDOW = True
DEFAULT_TLOG_REC_LIMIT_RATE = 16384
DEFAULT_TLOG_REC_LIMIT_BURST = 32768
DEFAULT_TLOG_REC_LIMIT_ACTION = "pass"
DEFAULT_TLOG_REC_WRITER = "file"
DEFAULT_TLOG_REC_JOURNAL_PRIORITY = "info"
DEFAULT_TLOG_REC_JOURNAL_AUGMENT = True
DEFAULT_TLOG_REC_SYSLOG_FACILITY = "authpriv"
DEFAULT_TLOG_REC_SYSLOG_PRIORITY = "info"

DEFAULT_TLOG_REC_SESSION_SHELL = "/bin/bash"
DEFAULT_TLOG_REC_SESSION_NOTICE = "ATTENTION: Your session is being recorded!"
DEFAULT_TLOG_REC_SESSION_WRITER = "journal"

DEFAULT_TLOG_PLAY_READER = "file"
DEFAULT_TLOG_PLAY_PERSIST = False
DEFAULT_TLOG_PLAY_LAX = False


class TlogPlayConfig:
    """Tlog Play configuration class"""
    def __init__(self, reader=DEFAULT_TLOG_PLAY_READER,
                 persist=DEFAULT_TLOG_PLAY_PERSIST,
                 lax=DEFAULT_TLOG_PLAY_LAX, file_reader_path=None,
                 es_baseurl=None,
                 es_query=None):
        self.reader = reader
        self.persist = persist
        self.lax = lax
        self.file_reader_path = file_reader_path
        self.es_baseurl = es_baseurl
        self.es_query = es_query

    def _setup_base_play_config(self):
        # unconditional tlog-play configuration options
        base_config = {
            "persist": self.persist,
            "lax": self.lax
        }

        return base_config

    def _setup_reader_config(self):
        if self.reader == "file":
            if self.file_reader_path is None:
                raise ValueError("file_reader_path must be provided")
            reader_config = {
                "reader": "file",
                "file": {
                    "path": self.file_reader_path
                },
            }
        elif self.reader == "es":
            if self.es_baseurl is None or self.es_query is None:
                raise ValueError("es_baseurl and es_query must be provided")
            reader_config = {
                "reader": "es",
                "es": {
                    "baseurl": self.es_baseurl,
                    "query": self.es_query
                },
            }
        else:
            raise ValueError("Unknown reader")

        return reader_config

    def _setup_config(self):
        config = self._setup_base_play_config()
        # append the reader configuration to the base configuration
        reader_config = self._setup_reader_config()
        config.update(reader_config)
        return config

    def generate_config(self, filename):
        """ Generate a configuration and write it to filename """
        tlog_play_config = self._setup_config()
        mkcfgfile(filename, json.dumps(tlog_play_config, indent=4))


class TlogRecConfig:
    """Tlog Rec configuration class"""
    def __init__(self, latency=DEFAULT_TLOG_REC_LATENCY,
                 payload=DEFAULT_TLOG_REC_PAYLOAD,
                 log_input=DEFAULT_TLOG_REC_LOG_INPUT,
                 log_output=DEFAULT_TLOG_REC_LOG_OUTPUT,
                 log_window=DEFAULT_TLOG_REC_LOG_WINDOW,
                 limit_rate=DEFAULT_TLOG_REC_LIMIT_RATE,
                 limit_burst=DEFAULT_TLOG_REC_LIMIT_BURST,
                 limit_action=DEFAULT_TLOG_REC_LIMIT_ACTION,
                 writer=DEFAULT_TLOG_REC_WRITER, file_writer_path=None,
                 journal_priority=DEFAULT_TLOG_REC_JOURNAL_PRIORITY,
                 journal_augment=DEFAULT_TLOG_REC_JOURNAL_AUGMENT,
                 syslog_facility=DEFAULT_TLOG_REC_SYSLOG_FACILITY,
                 syslog_priority=DEFAULT_TLOG_REC_SYSLOG_PRIORITY):
        self.latency = latency
        self.payload = payload
        self.log_input = log_input
        self.log_output = log_output
        self.log_window = log_window
        self.limit_rate = limit_rate
        self.limit_burst = limit_burst
        self.limit_action = limit_action
        self.writer = writer
        self.file_writer_path = file_writer_path
        self.journal_priority = journal_priority
        self.journal_augment = journal_augment
        self.syslog_facility = syslog_facility
        self.syslog_priority = syslog_priority

    def _setup_base_config(self):
        # unconditional tlog-rec configuration options
        base_config = {
            "latency": self.latency,
            "payload": self.payload,
            "log": {
                "input": self.log_input,
                "output": self.log_output,
                "window": self.log_window
            },
            "limit": {
                "rate": self.limit_rate,
                "burst": self.limit_burst,
                "action": self.limit_action
            },
        }

        return base_config

    def _setup_writer_config(self):
        # create the writer configuration based on input
        if self.writer == "file":
            if self.file_writer_path is None:
                raise ValueError("file_writer_path must be provided")
            writer_config = {
                "file": {
                    "path": self.file_writer_path
                },
                "writer": self.writer
            }
        elif self.writer == "journal":
            writer_config = {
                "journal": {
                    "priority": self.journal_priority,
                    "augment": self.journal_augment
                },
                "writer": self.writer
            }
        elif self.writer == "syslog":
            writer_config = {
                "syslog": {
                    "facility": self.syslog_facility,
                    "priority": self.syslog_priority
                },
                "writer": self.writer
            }
        else:
            raise ValueError("Unknown writer")

        return writer_config

    def _setup_config(self):
        config = self._setup_base_config()

        # append the writer configuration to the base configuration
        writer_config = self._setup_writer_config()
        config.update(writer_config)
        return config

    def generate_config(self, filename):
        """ Generate a configuration and write it to filename """

        tlog_rec_config = self._setup_config()
        mkcfgfile(filename, json.dumps(tlog_rec_config, indent=4))


class TlogRecSessionConfig(TlogRecConfig):
    """TlogPlaySession configuration class, child of TlogRecConfig"""
    def __init__(self, shell=DEFAULT_TLOG_REC_SESSION_SHELL,
                 notice=DEFAULT_TLOG_REC_SESSION_NOTICE,
                 latency=DEFAULT_TLOG_REC_LATENCY,
                 payload=DEFAULT_TLOG_REC_PAYLOAD,
                 log_input=DEFAULT_TLOG_REC_LOG_INPUT,
                 log_output=DEFAULT_TLOG_REC_LOG_OUTPUT,
                 log_window=DEFAULT_TLOG_REC_LOG_WINDOW,
                 limit_rate=DEFAULT_TLOG_REC_LIMIT_RATE,
                 limit_burst=DEFAULT_TLOG_REC_LIMIT_BURST,
                 limit_action=DEFAULT_TLOG_REC_LIMIT_ACTION,
                 writer=DEFAULT_TLOG_REC_SESSION_WRITER, file_writer_path=None,
                 journal_priority=DEFAULT_TLOG_REC_JOURNAL_PRIORITY,
                 journal_augment=DEFAULT_TLOG_REC_JOURNAL_AUGMENT,
                 syslog_facility=DEFAULT_TLOG_REC_SYSLOG_FACILITY,
                 syslog_priority=DEFAULT_TLOG_REC_SYSLOG_PRIORITY):
        self.shell = shell
        self.notice = notice
        super().__init__(latency, payload, log_input, log_output, log_window,
                         limit_rate, limit_burst, limit_action,
                         writer, file_writer_path, journal_priority,
                         journal_augment, syslog_facility,
                         syslog_priority)

    def _setup_base_session_config(self):
        # unconditional tlog-rec-session configuration options
        tlog_rec_session_config = {
            "shell": self.shell,
            "notice": self.notice,
        }

        return tlog_rec_session_config

    def generate_config(self, filename):
        """ Generate a configuration and write it to filename """

        # append TlogRecConfig configuration to our base configuration
        tlog_rec_session_config = self._setup_base_session_config()
        tlog_rec_config = super()._setup_config()
        tlog_rec_session_config.update(tlog_rec_config)

        mkcfgfile(filename, json.dumps(tlog_rec_session_config, indent=4))



""" tlog tests """
import os
import sys
import ast
import stat
import time
from tempfile import mkdtemp

import pexpect

import pytest

from misc import ssh_pexpect, journal_find_last, \
                 mklogfile, mkrecording, \
                 read_tlog_recording_file


class TestTlogPlay:
    """ Tests for tlog-play basic usage """
    user1 = 'tlitestlocaluser1'
    tempdir = mkdtemp(prefix='/tmp/TestTlogPlay.')
    os.chmod(tempdir, stat.S_IRWXU + stat.S_IRWXG + stat.S_IRWXO +
             stat.S_ISUID + stat.S_ISGID + stat.S_ISVTX)

    @pytest.mark.tier1
    def test_play_from_file(self):
        """
        Check tlog-play can playback session from file
        """
        logfile = mklogfile(self.tempdir)
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        mkrecording(shell, logfile)
        shell.sendline('tlog-play -i {}'.format(logfile))
        shell.expect('localhost')

    @pytest.mark.tier1
    def test_play_from_journal(self):
        """
        Check tlog-play can playback session from journal
        """
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        mkrecording(shell, sleep=4)
        shell.close()

        entry = journal_find_last()
        message = entry['MESSAGE']
        rec = ast.literal_eval(message)['rec']
        tlog_rec = 'TLOG_REC={}'.format(rec)
        cmd = 'tlog-play -r journal -M {}'.format(tlog_rec)
        shell2 = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        shell2.sendline(cmd)
        out = shell2.expect([pexpect.TIMEOUT, 'localhost'], timeout=10)
        assert out == 1

    def test_play_output_file(self):
        """
        Ensure tlog-play will redirect output to file
        """

        outputfile = mklogfile(self.tempdir)
        recordedtext = 'output_test'
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        shell.sendline(f'tlog-rec -w journal echo {recordedtext}')
        time.sleep(5)

        entry = journal_find_last()
        message = entry['MESSAGE']
        rec = ast.literal_eval(message)['rec']
        tlog_rec = 'TLOG_REC={}'.format(rec)

        playcmd = f'tlog-play -r journal -M {tlog_rec} > {outputfile}'
        shell2 = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        shell2.sendline(playcmd)
        time.sleep(5)

        with open(outputfile) as f:
            read_data = f.read()
            assert recordedtext in read_data
        f.closed

    def test_play_at_speed_x2(self):
        """
        Check tlog-play can playback session at 2x speed
        """
        logfile = mklogfile(self.tempdir)
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        mkrecording(shell, logfile, sleep=15)
        shell.close()

        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        time_start = time.time()
        opts = '-i {} --speed=2'.format(logfile)
        shell.sendline('tlog-play {}'.format(opts))
        shell.expect('localhost')
        time_stop = time.time()
        assert time_stop-time_start < 9

    def test_play_at_goto_points(self):
        """
        Check tlog-play can start playback session from goto points
        """
        logfile = mklogfile(self.tempdir)
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        shell.sendline('tlog-rec -o {}'.format(logfile))
        shell.sendline('echo start')
        time.sleep(5)
        shell.sendline('echo test_1')
        shell.expect('test_1')
        time.sleep(5)
        shell.sendline('echo test_2')
        shell.expect('test_2')
        time.sleep(5)
        shell.sendline('echo test_3')
        shell.expect('test_3')
        shell.sendline('echo end')
        shell.sendline('exit')
        shell.close()

        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        time_start = time.time()
        opts = '-i {} -g start'.format(logfile)
        shell.sendline('tlog-play {}'.format(opts))
        shell.expect('end')
        time_stop = time.time()
        assert time_stop-time_start > 15
        shell.close()

        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        time_start = time.time()
        opts = '-i {} -g end'.format(logfile)
        shell.sendline('tlog-play {}'.format(opts))
        shell.expect('end')
        time_stop = time.time()
        assert time_stop-time_start < 2
        shell.close()

        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        time_start = time.time()
        opts = '-i {} -g 11'.format(logfile)
        shell.sendline('tlog-play {}'.format(opts))
        shell.expect('end')
        time_stop = time.time()
        assert time_stop-time_start < 8
        shell.close()

    def test_play_follow_running_session(self):
        """
        Check tlog-play can follow a live running session
        """
        logfile = mklogfile(self.tempdir)
        shell1 = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        shell2 = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        shell2.logfile = sys.stdout
        shell1.sendline('tlog-rec -o {}'.format(logfile))
        time.sleep(1)
        shell2.sendline('tlog-play -i {} -f'.format(logfile))
        shell1.sendline('echo line1')
        shell2.expect('line1')
        time.sleep(2)
        shell1.sendline('echo line2')
        shell2.expect('line2')
        time.sleep(2)
        shell1.sendline('echo line3')
        shell2.expect('line3')
        shell1.close()
        shell2.close()


class TestTlogPlayControl:
    """ Tests for tlog-play running key controls usage """
    user1 = 'tlitestlocaluser1'
    tempdir = mkdtemp(prefix='/tmp/TestTlogPlay.')
    os.chmod(tempdir, stat.S_IRWXU + stat.S_IRWXG + stat.S_IRWXO +
             stat.S_ISUID + stat.S_ISGID + stat.S_ISVTX)
    ctl_log = mklogfile(tempdir)

    @classmethod
    def setup_class(cls):
        """ create sample recorded session for key control tests """
        shell = ssh_pexpect(cls.user1, 'Secret123', 'localhost')
        shell.sendline('tlog-rec -o {}'.format(cls.ctl_log))
        shell.sendline('echo start')
        time.sleep(10)
        shell.sendline('echo testime_stop')
        time.sleep(10)
        shell.sendline('echo test2')
        time.sleep(10)
        shell.sendline('echo test3')
        time.sleep(10)
        shell.sendline('echo end')
        shell.sendline('exit')
        shell.close()

    def test_play_key_ctl_goto_end(self):
        """
        Check tlog-play key control goes to end of session
        """
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        time_start = time.time()
        shell.sendline('tlog-play -i {}'.format(self.ctl_log))
        time.sleep(1)
        shell.sendline('G')
        shell.expect('end')
        time_stop = time.time()
        assert time_stop-time_start < 3
        shell.close()

    def test_play_key_ctl_next_packet(self):
        """
        Check tlog-play key control skips to next packet
        """
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        time_start = time.time()
        shell.sendline('tlog-play -i {}'.format(self.ctl_log))
        time.sleep(1)
        shell.sendline('.')
        time.sleep(1)
        shell.sendline('.')
        time.sleep(1)
        shell.sendline('.')
        time.sleep(1)
        shell.sendline('.')
        shell.expect('end')
        time_stop = time.time()
        assert time_stop-time_start < 5
        shell.close()

    def test_play_key_ctl_double_speed(self):
        """
        Check tlog-play key control doubles playback speed
        """
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        time_start = time.time()
        shell.sendline('tlog-play -i {}'.format(self.ctl_log))
        time.sleep(1)
        shell.sendline('}')
        time.sleep(1)
        shell.sendline('}')
        shell.expect('end')
        time_stop = time.time()
        assert time_stop-time_start < 15
        shell.close()

    def test_play_key_ctl_quit(self):
        """
        Check tlog-play key control quits playback
        """
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        time_start = time.time()
        shell.sendline('tlog-play -i {}'.format(self.ctl_log))
        time.sleep(1)
        shell.sendline('q')
        time_stop = time.time()
        assert time_stop-time_start < 2
        shell.close()

    def test_play_key_ctl_halve_speed(self):
        """
        Check tlog-play key control halves speed

        The double speed steps are there as necessary pre-req steps
        """
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        time_start = time.time()
        shell.sendline('tlog-play -i {}'.format(self.ctl_log))
        time.sleep(1)
        shell.sendline('}')
        time.sleep(1)
        shell.sendline('}')
        time.sleep(1)
        shell.sendline('{')
        time.sleep(1)
        shell.sendline('{')
        shell.expect('end', timeout=40)
        time_stop = time.time()
        assert time_stop-time_start > 30
        shell.close()

    def test_play_from_file_match(self):
        """
        File reader matching functionality test
        """
        logfile = mklogfile(self.tempdir)
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        shell.sendline('tlog-rec -o {} whoami'.format(logfile))
        time.sleep(5)
        shell.sendline('tlog-rec -o {} lsblk'.format(logfile))
        time.sleep(5)
        shell.sendline('tlog-rec -o {} ls -l /usr/bin'.format(logfile))
        time.sleep(5)

        rec_ids = []
        recording = read_tlog_recording_file(logfile)
        for msg in recording:
            rec = msg['rec']
            if rec not in rec_ids:
                rec_ids.append(rec)

        shell.sendline('tlog-play -i {} -m {}'.format(logfile, rec_ids[0]))
        shell.expect(self.user1)
        shell.sendline('tlog-play -i {} -m {}'.format(logfile, rec_ids[1]))
        shell.expect('MOUNTPOINT')
        shell.sendline('tlog-play -i {} -m {}'.format(logfile, rec_ids[2]))
        shell.expect('uname')



""" tlog tests """

import os
import ast
import stat
import time
import inspect
import json
from shutil import copyfile
from systemd import journal

import pexpect
from pexpect.pxssh import pxssh


def journal_find_last():
    """ Find the last TLOG_REC journal entry """
    j = journal.Reader()
    j.seek_tail()
    while True:
        entry = j.get_previous()

        if '_COMM' in entry:
            matchfield = '_COMM'
        elif 'SYSLOG_IDENTIFIER' in entry:
            matchfield = 'SYSLOG_IDENTIFIER'
        else:
            continue

        if 'tlog' in entry[matchfield]:
            return entry
        elif not entry:
            raise ValueError('Did not find TLOG_REC entry in journal')


def check_journal(pattern):
    """ Check that last journal entry contains pattern """
    time.sleep(1)
    for _ in range(0, 10):
        entry = journal_find_last()
        message = entry['MESSAGE']
        out_txt = ast.literal_eval(message)['out_txt']
        if pattern in out_txt:
            break
        else:
            time.sleep(5)
    assert pattern in out_txt


def check_outfile(pattern, filename, maxchecks=10):
    """ Check that file contains pattern """
    time.sleep(1)
    for _ in range(0, maxchecks):
        file1 = open(filename, 'r')
        content = file1.read()
        file1.close()
        if pattern in content:
            break
        else:
            time.sleep(5)
    assert pattern in content


def check_recording(shell, pattern, filename=None):
    """ Check that recording contains pattern """
    time.sleep(1)
    if filename is not None:
        cmd = 'tlog-play -i {}'.format(filename)
    else:
        entry = journal_find_last()
        message = entry['MESSAGE']
        rec = ast.literal_eval(message)['rec']
        tlog_rec = 'TLOG_REC={}'.format(rec)
        cmd = 'tlog-play -r journal -M {}'.format(tlog_rec)
    shell.sendline(cmd)
    out = shell.expect([pexpect.TIMEOUT, pattern], timeout=10)
    # print(shell.before)
    # print(shell.after)
    if out == 0:
        print('\ncheck_recording TIMEOUT')
    assert out == 1


def check_recording_missing(shell, pattern, filename=None):
    """ Check that recording does not contain pattern """
    time.sleep(1)
    if filename is not None:
        cmd = 'tlog-play -g end -i {}'.format(filename)
    else:
        entry = journal_find_last()
        message = entry['MESSAGE']
        rec = ast.literal_eval(message)['rec']
        tlog_rec = 'TLOG_REC={}'.format(rec)
        cmd = 'tlog-play -g end -r journal -M {}'.format(tlog_rec)
    shell.sendline(cmd)
    out = shell.expect([pexpect.TIMEOUT, pattern], timeout=5)
    if out == 1:
        print('\ncheck_recording_missing found: {}'.format(pattern))
    assert out == 0


def ssh_pexpect(username, password, hostname, encoding='utf-8'):
    """ Setup an ssh connection to remote host """
    ssh = pxssh(options={
        "StrictHostKeyChecking": "no",
        "UserKnownHostsFile": "/dev/null"
        }, encoding=encoding, codec_errors='replace')
    ssh.force_password = True
    ssh.login(hostname, username, password)
    ssh.sendline('echo loggedin')
    ssh.expect('loggedin')
    # ssh.logfile = sys.stdout
    return ssh


def mklogfile(directory, filename=None):
    """ Create temporary logfile """
    # if filename not given, we use the name of the calling frame
    if filename is None:
        filename = '{}.tlog'.format(inspect.stack()[1][3])
    logfile = '{}/{}'.format(directory, filename)
    if not os.path.isdir(directory):
        os.makedirs(directory)
        os.chmod(directory, stat.S_IRWXU + stat.S_IRWXG + stat.S_IRWXO +
                 stat.S_ISUID + stat.S_ISGID + stat.S_ISVTX)
    return logfile


def mkcfgfile(filename, content):
    """ Create config file from string content """
    bkup = '{}.origtest'.format(filename)
    content = '\n'.join([line.lstrip() for line in content.split('\n')])
    if os.path.isfile(filename) and not os.path.isfile(bkup):
        copyfile(filename, bkup)
    with open(filename, 'w') as out:
        out.write(content)


def mkrecording(shell, filename=None, sleep=3):
    """ Create a standard recording for tests to use """
    if filename is None:
        opts = '-w journal'
    else:
        opts = '-o {}'.format(filename)
    shell.sendline('tlog-rec {}'.format(opts))
    time.sleep(sleep)
    shell.sendline('id')
    shell.sendline('cat /etc/hosts')
    shell.sendline('exit')

def read_tlog_recording_file(recording_file):
    """ Read tlog message(s) from file. JSON Objects are
        decoded and returned into list of dictionaries.
    """
    messages = []

    with open(recording_file) as f:
        for line in f:
            obj = json.loads(line)
            messages.append(obj)

    return messages



""" tlog remote recording tests """
import os

import pexpect

import pytest


class TestTlogPlayRemote:
    """ Test tlog-play for remotely recorded journal sessions """
    bootid = '6211b985bd1e49908726bd24cd514099'
    cases = {'good': '600-b0b9',
             'bad-timing': '72d-df52',
             'missing-entry': '786-11109'}

    @classmethod
    def setup_class(cls):
        """
        Import remotely recorded sessions for test cases to the
        journal
        """
        tpath = os.path.dirname(os.path.abspath(__file__))
        shell = pexpect.spawn('/bin/bash', cwd=os.getcwd())
        remdir = '/var/log/journal/remote'
        imported = '{}/test-journals-imported'.format(remdir)
        jremote = '/usr/lib/systemd/systemd-journal-remote'
        finish = 'Finishing after writing'
        if os.path.isfile(imported):
            return
        if not os.path.isdir(remdir):
            os.makedirs(remdir, mode=755, exist_ok=True)
        # for jtype in cls.cases.keys():
        for jtype, _ in cls.cases.items():
            jfile = '{}/journal-{}.txt'.format(tpath, jtype)
            jremfile = '{}/remote-{}.journal'.format(remdir, jtype)
            cmd = 'cat {}|{} -o {} -'.format(jfile, jremote, jremfile)
            shell.sendline(cmd)
            out = shell.expect([pexpect.TIMEOUT, finish], timeout=10)
            if out == 0:
                print('\nTestTlogPlayRemote setup TIMEOUT.')
                print('\nThere was a problem importing journals')
            assert out == 1
        open(imported, 'a').close()
        shell.close()

    @pytest.mark.tier1
    def test_play_remote_good_session(self):
        """
        Check tlog-play can playback remote session imported to
        journal
        """
        pattern = 'logout'
        shell = pexpect.spawn('/bin/bash')
        tlog_rec = 'TLOG_REC={}-{}'.format(self.bootid,
                                           self.cases['good'])
        cmd = 'tlog-play -r journal -M {} -g end'.format(tlog_rec)
        shell.sendline(cmd)
        out = shell.expect([pexpect.TIMEOUT, pattern], timeout=10)
        if out == 0:
            print('test_play_remote_good_session TIMEOUT')
        assert out == 1
        shell.close()

    def test_play_remote_bad_timing(self):
        """
        Check tlog-play fails on remote session with invalid timing
        """
        pattern = 'invalid "timing"'
        shell = pexpect.spawn('/bin/bash')
        tlog_rec = 'TLOG_REC={}-{}'.format(self.bootid,
                                           self.cases['bad-timing'])
        cmd = 'tlog-play -r journal -M {} -g end'.format(tlog_rec)
        shell.sendline(cmd)
        out = shell.expect([pexpect.TIMEOUT, pattern], timeout=10)
        if out == 0:
            print('test_play_remote_bad_timing TIMEOUT')
        assert out == 1
        shell.close()

    def test_play_remote_missing_entry(self):
        """
        Check tlog-play fails on remote session missing message
        """
        pattern = 'ID is out of order'
        shell = pexpect.spawn('/bin/bash')
        tlog_rec = 'TLOG_REC={}-{}'.format(self.bootid,
                                           self.cases['missing-entry'])
        cmd = 'tlog-play -r journal -M {} -g end'.format(tlog_rec)
        shell.sendline(cmd)
        out = shell.expect([pexpect.TIMEOUT, pattern], timeout=10)
        if out == 0:
            print('test_play_remote_missing_entry TIMEOUT')
        assert out == 1
        shell.close()



""" tlog tests """
import os
import stat
from tempfile import mkdtemp

from misc import ssh_pexpect, \
                 mklogfile, check_recording, \
                 check_recording_missing, check_outfile


class TestTlogRecPerformanceOptions:
    """ Test performance related options for tlog-rec """
    user1 = 'tlitestlocaluser1'
    tempdir = mkdtemp(prefix='/tmp/TestTlogRec.')
    os.chmod(tempdir, stat.S_IRWXU + stat.S_IRWXG + stat.S_IRWXO +
             stat.S_ISUID + stat.S_ISGID + stat.S_ISVTX)

    def test_record_fast_input_with_latency(self):
        """
        Check tlog-rec caches data some time before logging
        """
        logfile = mklogfile(self.tempdir)
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        opts = '--latency=9'
        shell.sendline('tlog-rec {} '
                       '-o {} /bin/bash'.format(opts, logfile))
        for num in range(0, 200):
            shell.sendline('echo test_{}'.format(num))
        shell.sendline('exit')
        check_recording(shell, 'test_199', logfile)
        shell.close()

    def test_record_fast_input_with_payload(self):
        """
        Check tlog-rec limits output payload size
        """
        logfile = mklogfile(self.tempdir)
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        opts = '--payload=128'
        shell.sendline('tlog-rec {} '
                       '-o {} /bin/bash'.format(opts, logfile))
        for num in range(0, 200):
            shell.sendline('echo test_{}'.format(num))
        shell.sendline('exit')
        check_recording(shell, 'test_199', logfile)
        shell.close()

    def test_record_fast_input_with_limit_rate(self):
        """
        Check tlog-rec records session with limit-rate argument
        """
        logfile = mklogfile(self.tempdir)
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        opts = '--limit-rate=10'
        shell.sendline('tlog-rec {} '
                       '-o {} /bin/bash'.format(opts, logfile))
        for num in range(0, 200):
            shell.sendline('echo test_{}'.format(num))
        shell.sendline('exit')
        shell.close()

    def test_record_fast_input_with_limit_burst(self):
        """
        Check tlog-rec allows limited burst of fast output
        """
        logfile = mklogfile(self.tempdir)
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        opts = '--limit-rate=10 --limit-burst=100'
        shell.sendline('tlog-rec {} '
                       '-o {} /bin/bash'.format(opts, logfile))
        for num in range(0, 200):
            shell.sendline('echo test_{}'.format(num))
        shell.sendline('exit')
        shell.close()

    def test_record_fast_input_with_limit_action_drop(self):
        """
        Check tlog-rec drops output when logging limit reached
        """
        logfile = mklogfile(self.tempdir)
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        opts = '--limit-rate=10 --limit-action=drop'
        cmd = 'cat /usr/share/dict/linux.words'
        shell.sendline('tlog-rec {} '
                       '-o {} {}'.format(opts, logfile, cmd))
        shell.close()
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        check_recording_missing(shell, 'Byronite', logfile)
        check_recording_missing(shell, 'zygote', logfile)

    def test_record_fast_input_with_limit_action_delay(self):
        """
        Check tlog-rec delays recording when logging limit reached
        """
        logfile = mklogfile(self.tempdir)
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        opts = '--limit-rate=10 --limit-action=delay'
        shell.sendline('tlog-rec {} '
                       '-o {} /bin/bash'.format(opts, logfile))
        for num in range(0, 200):
            shell.sendline('echo test_{}'.format(num))
        check_outfile('test_199', logfile, maxchecks=100)
        shell.sendline('exit')
        check_recording(shell, 'test_199', logfile)
        shell.close()

    def test_record_fast_input_with_limit_action_pass(self):
        """
        Check tlog-rec ignores logging limits
        """
        logfile = mklogfile(self.tempdir)
        shell = ssh_pexpect(self.user1, 'Secret123', 'localhost')
        opts = '--limit-rate=10 --limit-action=pass'
        shell.sendline('tlog-rec {} '
                       '-o {} /bin/bash'.format(opts, logfile))
        for num in range(0, 200):
            shell.sendline('echo test_{}'.format(num))
        shell.sendline('exit')
        check_recording(shell, 'test_199', logfile)
        shell.close()
